@@ -1,6 +1,6 @@
 /*  bam_tview.c -- tview subcommand.
 
-    Copyright (C) 2008-2014 Genome Research Ltd.
+    Copyright (C) 2008-2016 Genome Research Ltd.
     Portions copyright (C) 2013 Pierre Lindenbaum, Institut du Thorax, INSERM U1087, Université de Nantes.
 
     Author: Heng Li <lh3@sanger.ac.uk>
@@ -23,33 +23,16 @@ LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 DEALINGS IN THE SOFTWARE.  */
 
+#include <config.h>
+
 #include <regex.h>
 #include <assert.h>
 #include "bam_tview.h"
 #include <htslib/faidx.h>
 #include <htslib/sam.h>
 #include <htslib/bgzf.h>
-
-/*! @typedef
- @abstract      Type of function to be called by sam_fetch().
- @param  b     the alignment
- @param  data  user provided data
- */
-typedef int (*sam_fetch_f)(const bam1_t *b, void *data);
-
-int sam_fetch(samFile *fp, const hts_idx_t *idx, int tid, int beg, int end, void *data, sam_fetch_f func)
-{
-    int ret;
-    hts_itr_t* iter;
-    bam1_t* b = bam_init1();
-    iter = sam_itr_queryi(idx, tid, beg, end);
-    while ((ret = sam_itr_next(fp, iter, b)) >= 0) func(b, data);
-    hts_itr_destroy(iter);
-    bam_destroy1(b);
-    return (ret == -1)? 0 : ret;
-}
-
-
+#include "samtools.h"
+#include "sam_opts.h"
 
 khash_t(kh_rg)* get_rg_sample(const char* header, const char* sample)
 {
@@ -74,7 +57,8 @@ khash_t(kh_rg)* get_rg_sample(const char* header, const char* sample)
     return rg_hash;
 }
 
-int base_tv_init(tview_t* tv, const char *fn, const char *fn_fa, const char *samples)
+int base_tv_init(tview_t* tv, const char *fn, const char *fn_fa,
+                 const char *samples, const htsFormat *fmt)
 {
     assert(tv!=NULL);
     assert(fn!=NULL);
@@ -82,10 +66,10 @@ int base_tv_init(tview_t* tv, const char *fn, const char *fn_fa, const char *sam
     tv->color_for = TV_COLOR_MAPQ;
     tv->is_dot = 1;
 
-    tv->fp = sam_open(fn, "r");
+    tv->fp = sam_open_format(fn, "r", fmt);
     if(tv->fp == NULL)
     {
-        fprintf(stderr,"sam_open %s. %s\n", fn,fn_fa);
+        print_error_errno("tview", "can't open \"%s\"", fn);
         exit(EXIT_FAILURE);
     }
     // TODO bgzf_set_cache_size(tv->fp->fp.bgzf, 8 * 1024 *1024);
@@ -94,13 +78,13 @@ int base_tv_init(tview_t* tv, const char *fn, const char *fn_fa, const char *sam
     tv->header = sam_hdr_read(tv->fp);
     if(tv->header == NULL)
     {
-        fprintf(stderr,"Cannot read '%s'.\n", fn);
+        print_error("tview", "cannot read \"%s\"", fn);
         exit(EXIT_FAILURE);
     }
     tv->idx = sam_index_load(tv->fp, fn);
     if (tv->idx == NULL)
     {
-        fprintf(stderr,"Cannot read index for '%s'.\n", fn);
+        print_error("tview", "cannot read index for \"%s\"", fn);
         exit(EXIT_FAILURE);
     }
     tv->lplbuf = bam_lplbuf_init(tv_pl_func, tv);
@@ -132,7 +116,6 @@ void base_tv_destroy(tview_t* tv)
 
 int tv_pl_func(uint32_t tid, uint32_t pos, int n, const bam_pileup1_t *pl, void *data)
 {
-    extern const char bam_nt16_nt4_table[];
     tview_t *tv = (tview_t*)data;
     int i, j, c, rb, attr, max_ins = 0;
     uint32_t call = 0;
@@ -231,7 +214,7 @@ int tv_pl_func(uint32_t tid, uint32_t pos, int n, const bam_pileup1_t *pl, void
                     if (x > 4) x = 4;
                     attr |= tv->my_colorpair(tv,x);
                 } else if (tv->color_for == TV_COLOR_NUCL) {
-                    x = bam_nt16_nt4_table[bam_seqi(bam_get_seq(p->b), p->qpos)] + 5;
+                    x = seq_nt16_int[bam_seqi(bam_get_seq(p->b), p->qpos)] + 5;
                     attr |= tv->my_colorpair(tv,x);
                 } else if(tv->color_for == TV_COLOR_COL) {
                     x = 0;
@@ -241,7 +224,7 @@ int tv_pl_func(uint32_t tid, uint32_t pos, int n, const bam_pileup1_t *pl, void
                         case '2': x = 2; break;
                         case '3': x = 3; break;
                         case '4': x = 4; break;
-                        default: x = bam_nt16_nt4_table[bam_seqi(bam_get_seq(p->b), p->qpos)]; break;
+                        default: x = seq_nt16_int[bam_seqi(bam_get_seq(p->b), p->qpos)]; break;
                     }
                     x+=5;
                     attr |= tv->my_colorpair(tv,x);
@@ -272,9 +255,8 @@ int tv_pl_func(uint32_t tid, uint32_t pos, int n, const bam_pileup1_t *pl, void
 
 
 
-int tv_fetch_func(const bam1_t *b, void *data)
+static int tv_push_aln(const bam1_t *b, tview_t *tv)
 {
-    tview_t *tv = (tview_t*)data;
     /* If we are restricted to specific readgroups check RG is in the list */
     if ( tv->rg_hash )
     {
@@ -322,7 +304,11 @@ int base_draw_aln(tview_t *tv, int tid, int pos)
     }
     // draw aln
     bam_lplbuf_reset(tv->lplbuf);
-    sam_fetch(tv->fp, tv->idx, tv->curr_tid, tv->left_pos, tv->left_pos + tv->mcol, tv, tv_fetch_func);
+    hts_itr_t *iter = sam_itr_queryi(tv->idx, tv->curr_tid, tv->left_pos, tv->left_pos + tv->mcol);
+    bam1_t *b = bam_init1();
+    while (sam_itr_next(tv->fp, iter, b) >= 0) tv_push_aln(b, tv);
+    bam_destroy1(b);
+    hts_itr_destroy(iter);
     bam_lplbuf_push(0, tv->lplbuf);
 
     while (tv->ccol < tv->mcol) {
@@ -347,6 +333,7 @@ static void error(const char *format, ...)
 "   -d display      output as (H)tml or (C)urses or (T)ext \n"
 "   -p chr:pos      go directly to this position\n"
 "   -s STR          display only reads from this sample or group\n");
+        sam_global_opt_help(stderr, "-.--.-");
     }
     else
     {
@@ -359,17 +346,27 @@ static void error(const char *format, ...)
 }
 
 enum dipsay_mode {display_ncurses,display_html,display_text};
-extern tview_t* curses_tv_init(const char *fn, const char *fn_fa, const char *samples);
-extern tview_t* html_tv_init(const char *fn, const char *fn_fa, const char *samples);
-extern tview_t* text_tv_init(const char *fn, const char *fn_fa, const char *samples);
+extern tview_t* curses_tv_init(const char *fn, const char *fn_fa,
+                               const char *samples, const htsFormat *fmt);
+extern tview_t* html_tv_init(const char *fn, const char *fn_fa,
+                             const char *samples, const htsFormat *fmt);
+extern tview_t* text_tv_init(const char *fn, const char *fn_fa,
+                             const char *samples, const htsFormat *fmt);
 
 int bam_tview_main(int argc, char *argv[])
 {
     int view_mode=display_ncurses;
     tview_t* tv=NULL;
-    char *samples=NULL, *position=NULL;
+    char *samples=NULL, *position=NULL, *ref;
     int c;
-    while ((c = getopt(argc, argv, "s:p:d:")) >= 0) {
+
+    sam_global_args ga = SAM_GLOBAL_ARGS_INIT;
+    static const struct option lopts[] = {
+        SAM_OPT_GLOBAL_OPTIONS('-', 0, '-', '-', 0, '-'),
+        { NULL, 0, NULL, 0 }
+    };
+
+    while ((c = getopt_long(argc, argv, "s:p:d:", lopts, NULL)) >= 0) {
         switch (c) {
             case 's': samples=optarg; break;
             case 'p': position=optarg; break;
@@ -384,28 +381,28 @@ int bam_tview_main(int argc, char *argv[])
                 }
                 break;
             }
-            default: error(NULL);
+            default:  if (parse_sam_global_opt(c, optarg, lopts, &ga) == 0) break;
+                      /* else fall-through */
+            case '?': error(NULL);
         }
     }
     if (argc==optind) error(NULL);
 
+    ref = (optind+1>=argc)? ga.reference : argv[optind+1];
+
     switch(view_mode)
     {
         case display_ncurses:
-            {
-            tv = curses_tv_init(argv[optind], (optind+1>=argc)? 0 : argv[optind+1], samples);
+            tv = curses_tv_init(argv[optind], ref, samples, &ga.in);
             break;
-            }
+
         case display_text:
-            {
-            tv = text_tv_init(argv[optind], (optind+1>=argc)? 0 : argv[optind+1], samples);
+            tv = text_tv_init(argv[optind], ref, samples, &ga.in);
             break;
-            }
+
         case display_html:
-            {
-            tv = html_tv_init(argv[optind], (optind+1>=argc)? 0 : argv[optind+1], samples);
+            tv = html_tv_init(argv[optind], ref, samples, &ga.in);
             break;
-            }
     }
     if (tv==NULL)
     {
@@ -416,7 +413,9 @@ int bam_tview_main(int argc, char *argv[])
     if ( position )
     {
         int tid, beg, end;
-        *(char *)hts_parse_reg(position, &beg, &end) = '\0';
+        char *name_lim = (char *) hts_parse_reg(position, &beg, &end);
+        if (name_lim) *name_lim = '\0';
+        else beg = 0; // region parsing failed, but possibly a seq named "foo:a"
         tid = bam_name2id(tv->header, position);
         if (tid >= 0) { tv->curr_tid = tid; tv->left_pos = beg; }
     }
