@@ -40,10 +40,9 @@ sub blast2sam {
   $dummy_score = defined($opts{d});
   @sam = (); @sam[0,4,6..8,10] = ('', 255, '*', 0, 0, '*');
   while (<>) {
-    if ((@cigar || @cmaux) && (/^Query=/ || /Score =.*bits.*Expect/ || /^>\S+/)) { # print
+    if (@cigar && (/^Query=/ || /Score =.*bits.*Expect/ || /^>\S+/)) { # print
       &blast_print_sam(\@sam, \@cigar, \@cmaux, $qlen - $qend);
       @cigar = ();
-      @cmaux = ();
     }
     if (/^Query=\s(\S+)/) {
       $sam[2] = undef;
@@ -176,6 +175,4 @@ samtools view -hT your_ref.fasta your_file.sam > your_file_with_header.sam
 
 =back
 
-Note that queries that have no alignments will not be output (as unmapped).
-
 =cut
